//#include "interface_to_yices.c"
//#include "interface_to_lp_solve.c"
#include "interface_to_cplex.c"
#ifdef ZCHAFF
	#include "zchaff/SAT.h"
#endif
#include "general.h"

/*	This file is used to rule out deadlocks found by the WHS tool which are actually illegal because they can only be formed by intersecting worms.
	We enumerate all routing paths that can be established in the 3 and 4 marked channels.
	We then call an lp solver to find a set of paths P such that:
	1.) all the paths in P are pairwise disjoint
	2.) all the neighbors of the head of each path in P are included in some path in P
	3.) P is non empty
	4.) P is the smallest set of paths satisfying 1.) to 3.)
*/

//extern bool MIN_DL_NOT_NEC;
extern bool VERBOSE;
extern bool SMT;
extern bool LOCAL_MINIMA;
extern bool MINIMIZE_RESOURCES;

extern const int RET_DLF;
extern const int RET_DL_MIN;
extern const int RET_DL_NOT_MIN;
extern const int RET_ERROR;



int member(int a, const int *y, int size_y) {
	for(int j = size_y-1; j >= 0; --j)
		if (a == y[j])
			return true;
    return false;
}



/*	Generates all routing paths that can be established fopr destination `d' from `curr' to some channel that is marked 3 and for which destination `d' does not lead to
	2 marked neighbors.
*/
dyn_int_List* enum_paths (int curr, int d, Graph* g, Marks* marks, bool* visited, const Params* params, ResourceList* faulty_resources) {
	if (!visited[curr]) {
		if (marks[curr].marking == 2 || inst_convert_to_resource(curr,params).p == L)
			return NULL;

		visited[curr] = true;

		int i,j,k,ret1_size;
		dyn_int_List* ret = (dyn_int_List*) malloc(2 * sizeof(dyn_int_List));
		for (i=0;i<2;i++) {
			ret[i].num_elts = 0;
			ret[i].num_allocated = 0;
			ret[i].elts = NULL;
		}
		int ret_index = 0, ret_size = 2;

		
		Procnode dest_procnode = inst_convert_to_procnode(d,params);
		ResourceList* neighbors = inst_routing(inst_convert_to_resource(curr,params),&dest_procnode,params,faulty_resources);
		if (marks[curr].marking == 3) {
			i = 0;
			bool blocked = neighbors->at(i).valid();
			while (neighbors->at(i).valid()&&blocked)
				blocked = marks[inst_convert_resource_to_int(neighbors->at(i++),params)].marking != 2;
			if (blocked) {
				// Found a 3-marked channel, so there is at least one path
				ret[0].num_allocated = 2;
				ret[0].elts = (int*) malloc(2 * sizeof(int));
				ret[0].elts[ret[0].num_elts++] = curr; 
				ret_index++;
			}
		}
		/* Adding all paths from all neighbors */
		i = 0;
		while (neighbors->at(i).valid()) {
            dyn_int_List* ret1 = enum_paths(inst_convert_resource_to_int(neighbors->at(i++),params), d, g, marks, visited, params, faulty_resources);

			ret1_size=0, j=0;
			while(ret1 != NULL && ret1[j++].elts != NULL)
				ret1_size++;
			if (ret1_size != 0) {
				void* _tmp = realloc(ret, (ret_size+ret1_size+1) * sizeof(dyn_int_List));
				assert(_tmp != NULL);
				ret = (dyn_int_List*) _tmp;
				for (j=ret_size;j<ret_size+ret1_size+1;j++) {
					ret[j].num_elts = 0;
					ret[j].num_allocated = 0;
					ret[j].elts = NULL;
				}
				ret_size += ret1_size+1;

				j=0;
				while(ret1 != NULL && ret1[j].elts != NULL) {
					assert(ret_index < ret_size-1);
assert(ret[ret_index].num_elts  == 0);
					ret[ret_index].num_allocated = ret[ret_index].num_elts + ret1[j].num_elts+1;
					ret[ret_index].elts = (int*) malloc(ret[ret_index].num_allocated * sizeof(int));
					ret[ret_index].elts[ret[ret_index].num_elts++] = curr;
					for (k=0;k<ret1[j].num_elts;k++)
						ret[ret_index].elts[ret[ret_index].num_elts++] = ret1[j].elts[k];
					ret_index++;
					free(ret1[j].elts);
					j++;
				}
			}
			free(ret1);
        }
/*if(ret[0].elts == NULL) {
		if (marks[curr].marking == 3) {
			i = 0;
			bool blocked = neighbors->at(i).valid();
			while (neighbors->at(i).valid()&&blocked)
				blocked = marks[inst_convert_resource_to_int(neighbors->at(i++),params)].marking != 2;
			if (blocked) {
*/				/* Found a 3-marked channel, so there is at least one path */
/*				ret[0].num_allocated = 2;
				ret[0].elts = (int*) malloc(2 * sizeof(int));
				ret[0].elts[ret[0].num_elts++] = curr; 
				ret_index++;
			}
		}
}*/
		delete neighbors;
		visited[curr] = false;
		return ret;
	}
	return NULL;
}


bool equivalent_paths(dyn_int_List p1, Procnode d1, dyn_int_List p2, Procnode d2, const Params* params, ResourceList* faulty_resources) {
	int i,j;
	bool found;
	bool ret = p1.num_elts==p2.num_elts;
	for (i=0;i<p1.num_elts&&ret;i++)
		ret = p1.elts[i] == p2.elts[i];
	if (ret) {
		ResourceList* neighbors_1 = inst_routing(inst_convert_to_resource(p1.elts[p1.num_elts-1],params),&d1,params,faulty_resources);
		ResourceList* neighbors_2 = inst_routing(inst_convert_to_resource(p2.elts[p2.num_elts-1],params),&d2,params,faulty_resources);
		i = 0;
		while (neighbors_1->at(i).valid()&&ret) {
			j=0;
			found = false;
			while (neighbors_2->at(j).valid()&&!found)
				found = equal_resources(neighbors_1->at(i),neighbors_2->at(j++));
			ret = found;
			i++;
		}
		i=0;
		while (neighbors_2->at(i).valid()&&ret) {
			j=0;
			found = false;
			while (neighbors_1->at(j).valid()&&!found)
				found = equal_resources(neighbors_2->at(i),neighbors_1->at(j++));
			ret = found;
			i++;
		}
		delete neighbors_1;
		delete neighbors_2;
	}
	return ret;
}

void print_worm(dyn_int_List path, int d, const Params* params, ResourceList* faulty_resources) {
	int i;
	printf("Worm with destination ");
	Procnode dest_procnode = inst_convert_to_procnode(d,params);
	print_procnode(dest_procnode,params);
	printf(" consisting of channels: [");
	for (i=0;i<path.num_elts;i++)
		std::cout << inst_convert_to_resource(path.elts[i],params) << (i<path.num_elts-1 ? ", " : "");
	printf("]");
	printf(" routed towards [");
	ResourceList* neighbors = inst_routing(inst_convert_to_resource(path.elts[path.num_elts-1],params),&dest_procnode,params,faulty_resources);
	i = 0;
	int num_of_neighbors = 0;
	while (neighbors->at(i++).valid())
		num_of_neighbors++;
	i=0;
	while (neighbors->at(i).valid()) {
		std::cout << neighbors->at(i) << (i == num_of_neighbors-1 ? "" : ",");
		i++;
	}
	delete neighbors;
	printf("]\n");
}

void remove_equivalent_paths(dyn_int_List* set_of_d_paths, int* d_paths_dests, int set_of_d_paths_offset, int* set_of_d_paths_size, const Params* params, ResourceList* faulty_resources) {
	int i,j,k;
	for	(i=set_of_d_paths_offset;i<*set_of_d_paths_size-1;i++) {
		if (d_paths_dests[i] != -1) {
			for	(j=i+1;j<*set_of_d_paths_size;j++) {
				if (equivalent_paths(set_of_d_paths[i], inst_convert_to_procnode(d_paths_dests[i],params), set_of_d_paths[j], inst_convert_to_procnode(d_paths_dests[j],params), params, faulty_resources)) {
					free(set_of_d_paths[j].elts);
					for	(k=j;k<*set_of_d_paths_size;k++) {
						set_of_d_paths[k] = set_of_d_paths[k+1];
						d_paths_dests[k] = d_paths_dests[k+1];
					}
					j--;
					(*set_of_d_paths_size)--;
				}
			}
		}
	}
}


void remove_small_paths(dyn_int_List* set_of_d_paths, int* d_paths_dests, int* set_of_d_paths_size, const Params* params, ResourceList* faulty_resources) {
	int i,k;
	for	(i=0;i<*set_of_d_paths_size;i++) {
		if (set_of_d_paths[i].num_elts != 4) {
			free(set_of_d_paths[i].elts);
			for	(k=i;k<*set_of_d_paths_size;k++) {
				set_of_d_paths[k] = set_of_d_paths[k+1];
				d_paths_dests[k] = d_paths_dests[k+1];
			}
			i--;
			(*set_of_d_paths_size)--;
		}
	}
}



struct garbage_bool_int {
	int** ints;
	bool** bools;
	int num_of_ints;
	int num_of_bools;
	int num_of_ints_allocated;
	int num_of_bools_allocated;
};

void add_int_to_garbage_bool_int(int* ints, struct garbage_bool_int* g) {
	if (g->num_of_ints >= g->num_of_ints_allocated) {
		g->num_of_ints_allocated *= 2;
		g->ints = (int**) realloc(g->ints, g->num_of_ints_allocated * sizeof(int*));
		assert(g->ints != NULL);
	}
	g->ints[g->num_of_ints++] = ints;
}
void add_bool_to_garbage_bool_int(bool* bools, struct garbage_bool_int* g) {
	if (g->num_of_bools >= g->num_of_bools_allocated) {
		g->num_of_bools_allocated *= 2;
		g->bools = (bool**) realloc(g->bools, g->num_of_bools_allocated * sizeof(bool*));
		assert(g->bools != NULL);
	}	
	g->bools[g->num_of_bools++] = bools;
}
void free_garbage_bool_int(struct garbage_bool_int* g) {
	int i;
	for (i=0;i<g->num_of_ints;i++)
		free(g->ints[i]);
	for (i=0;i<g->num_of_bools;i++)
		free(g->bools[i]);
	free(g->ints);
	free(g->bools);
	g->ints = NULL;
	g->bools = NULL;
}

/* creates a copy of the list */
int* cpy_to_int_array(const int *ints, int n, struct garbage_bool_int* g) {
    int i = 0;
    int *ret_elts = (int*) malloc (n * sizeof(int));
	add_int_to_garbage_bool_int(ret_elts, g);
    for (i=0; i<n; i++)
        ret_elts[i] = ints[i];
    return ret_elts;
}
/* creates a copy of the list */
bool* cpy_to_bool_array(const bool *bools, int n, struct garbage_bool_int* g) {
    int i = 0;
    bool *ret_elts = (bool*) malloc (n * sizeof(bool));
	add_bool_to_garbage_bool_int(ret_elts, g);
    for (i=0; i<n; i++)
        ret_elts[i] = bools[i];
    return ret_elts;
}


int interaction_with_solver(bool min_dl_not_nec_local, bool local_minimum, dyn_int_List* set_of_d_paths, int * d_paths_dests, int set_of_d_paths_index, Graph* g, Marks* marks, ResourceList* faulty_resources, const Params* params) {
	/* Creating the constraints */
	#ifdef ZCHAFF
		SAT_Manager sat;
	#endif
	struct constraint objective;
	int cs_size = 500;
	struct constraint * cs = (struct constraint*) malloc (cs_size * sizeof(struct constraint));
	int cs_index = 0;
	int r,i,j,num_of_resources = bitsize_of_resources(params);
	int** garbage_bool_int_ints = (int**) malloc (cs_size * sizeof(int*));
	bool** garbage_bool_int_bools = (bool**) malloc (cs_size * sizeof(bool*));
	struct garbage_bool_int garbage = {garbage_bool_int_ints, garbage_bool_int_bools, 0, 0,cs_size,cs_size};


	int path,path1;
	/* Constraint to be minimized */
	if (!min_dl_not_nec_local) {
		//LP
		if (true) {
			// Minimize number of worms
			objective.lhs = (int*) malloc (set_of_d_paths_index * sizeof(int));
			objective.pos = (int*) malloc (set_of_d_paths_index * sizeof(int));
			objective.lhs_size = 0;
			for (path=0;path<set_of_d_paths_index;path++) {
				objective.lhs[objective.lhs_size] = path;
				objective.pos[objective.lhs_size++] = 1;
			}
			objective.constr_type = OBJ;
		}
		else {
			// Minimize number of channels
			objective.lhs = (int*) malloc (num_of_resources * sizeof(int));
			objective.pos = (int*) malloc (num_of_resources * sizeof(int));
			objective.lhs_size = 0;
			for (r=0; r<num_of_resources; r++) {
        		if ((marks[r].marking == 3 || marks[r].marking == 4) && inst_convert_to_resource(r,params).p != L) {
					objective.lhs[objective.lhs_size] = set_of_d_paths_index+r;
					objective.pos[objective.lhs_size++] = 1;
				}
			}
			objective.constr_type = OBJ;
		}
	}
	else if (SMT) {
		//SMT
	}
	else {
		//SAT
		#ifdef ZCHAFF
		if (VERBOSE) printf("Solve the following CNF:\np cnf %i NUM_OF_LINES\n", set_of_d_paths_index);
		sat = SAT_InitManager();
		SAT_SetNumVariables(sat, set_of_d_paths_index+num_of_resources);
		#else
			printf("Install zchaff to make use of this option. Please contact Freek Verbeek for questions.\n");
			assert(false);
		#endif
	}
	/* Any channel can be in one path only */
	int lhs_index;
	int* lhs = (int*) malloc (set_of_d_paths_index * sizeof(int));
	int* pos = !min_dl_not_nec_local || SMT ? (int*) malloc (set_of_d_paths_index * sizeof(int)) : NULL;
	for (r=0; r<num_of_resources; r++) {
if (r%100==0) printf("%i/%i\n", r,num_of_resources);
        if ((marks[r].marking == 3 || marks[r].marking == 4) && inst_convert_to_resource(r,params).p != L) {
			int lhs_index = 0;
			if (!min_dl_not_nec_local || SMT) {
				//LP,SMT					
				pos[lhs_index]   = -1;
				lhs[lhs_index++] = set_of_d_paths_index+r;
				for (path1=0;path1<set_of_d_paths_index;path1++) {
					if (member(r, set_of_d_paths[path1].elts, set_of_d_paths[path1].num_elts)) {
						pos[lhs_index]   = 1;
						lhs[lhs_index++] = path1;
					}
				}
				if (lhs_index > 1) {
					int rhs = 0;
					struct constraint c = {cpy_to_int_array(lhs,lhs_index,&garbage), cpy_to_int_array(pos,lhs_index,&garbage), lhs_index, rhs, EQ};
					if (cs_index >= cs_size) {
						cs_size *= 2;
						cs = (struct constraint*) realloc(cs, cs_size * sizeof(struct constraint));
						assert(cs != NULL);
					}
					cs[cs_index++] = c;
				}
			}
			else {
				//SAT
				#ifdef ZCHAFF
				for (path1=0;path1<set_of_d_paths_index;path1++)
					if (member(r, set_of_d_paths[path1].elts, set_of_d_paths[path1].num_elts))
						lhs[lhs_index++] = path1;
assert(lhs_index>0);
int *clause = (int*) malloc((lhs_index+2)*sizeof(int));
clause[0] = (set_of_d_paths_index+r+1)*2+1;
for (i=0;i<lhs_index;i++)
	clause[i+1] =(lhs[i]+1)*2+0;
SAT_AddClause(sat, clause, lhs_index+1, 0);
cs_index++;
free(clause);
				if (lhs_index > 1) {
					for (i=0;i<lhs_index-1;i++) {
						for(j=i+1;j<lhs_index;j++) {
							if (VERBOSE) printf("-%i -%i 0\n", lhs[i]+1, lhs[j]+1);
							int clause[2];
							clause[0] = (lhs[i]+1)*2+1;
							clause[1] = (lhs[j]+1)*2+1;
							assert(clause[0]!=clause[1]);
							SAT_AddClause(sat, clause, 2, 0);
							cs_index++;
						}
					}
				}
				#endif
			}
		}
	}

	/* For any path, the head of the path must be blocked */
	for (path=0;path<set_of_d_paths_index;path++) {
if (path%1000==0) printf("%i/%i\n", path,set_of_d_paths_index);
		Procnode dest_procnode = inst_convert_to_procnode(d_paths_dests[path],params);
		ResourceList* neighbors = inst_routing(inst_convert_to_resource(set_of_d_paths[path].elts[set_of_d_paths[path].num_elts-1],params),&dest_procnode,params,faulty_resources);
		if (!min_dl_not_nec_local || SMT) {
			//LP,SMT
			lhs_index = 0;
			pos[lhs_index]   = -1;
			lhs[lhs_index++] = path;
			i = 0;
			while (neighbors->at(i).valid()) {
				if (!is_faulty_resource(neighbors->at(i), faulty_resources)) {
					pos[lhs_index] = 1;
					lhs[lhs_index++] = set_of_d_paths_index + inst_convert_resource_to_int(neighbors->at(i),params);
					//if (member(inst_convert_resource_to_int(neighbors->at(i),params), set_of_d_paths[path].elts, set_of_d_paths[path].num_elts))
					//	do_not_add_constraint = true;
					//assert (!member(inst_convert_resource_to_int(neighbors->at(i),params), set_of_d_paths[path].elts, set_of_d_paths[path].num_elts));
				}
				assert(lhs_index > 1);
				int rhs = 0;
				struct constraint c = {cpy_to_int_array(lhs,lhs_index,&garbage), cpy_to_int_array(pos,lhs_index,&garbage), lhs_index, rhs, GE};
				if (cs_index >= cs_size) {
					cs_size *= 2;
					cs = (struct constraint*) realloc(cs, cs_size * sizeof(struct constraint));
					assert(cs != NULL);
				}
				lhs_index = 1;
				cs[cs_index++] = c;
				i++;
			}
		}
		else {
			//SAT
			#ifdef ZCHAFF
			i = 0;
			while (neighbors->at(i).valid()) {
				if (!is_faulty_resource(neighbors->at(i), faulty_resources)) {
					lhs_index = 0;

lhs[lhs_index++] = (path+1)*2+1;
lhs[lhs_index++] = (set_of_d_paths_index + 1 + inst_convert_resource_to_int(neighbors->at(i),params))*2+0;
SAT_AddClause(sat, cpy_to_int_array(lhs,lhs_index,&garbage), lhs_index, 0);
cs_index++;
/*
					do_not_add_constraint = false;
					lhs[lhs_index++] = (path+1)*2+1;
					for (path1=0;path1<set_of_d_paths_index&&!do_not_add_constraint;path1++) {
						if (member(inst_convert_resource_to_int(neighbors->at(i),params), set_of_d_paths[path1].elts, set_of_d_paths[path1].num_elts)) {
							if (path1==path)
								do_not_add_constraint = true;
							else
								lhs[lhs_index++] = (path1+1)*2+0;
						}
					}
					if (!do_not_add_constraint) {
						if (VERBOSE) {
							printf("-%i ", path+1);
							for (path1=0;path1<set_of_d_paths_index;path1++)
								if (member(inst_convert_resource_to_int(neighbors->at(i),params), set_of_d_paths[path1].elts, set_of_d_paths[path1].num_elts))
									printf("%i ", path1+1);
							printf("0\n");
						}
if (lhs_index<=1){
print_worm(set_of_d_paths[path], d_paths_dests[path], params, faulty_resources);
std::cout << "Neighbor = " << neighbors->at(i) << "\n";
}
						assert(lhs_index > 1);
						SAT_AddClause(sat, cpy_to_int_array(lhs,lhs_index,&garbage), lhs_index, 0);
						cs_index++;
					}
*/
				}
				i++;
			}
			#endif
		}
		delete neighbors;
	}
	free(lhs);
	free(pos);
	/* The solution must be non-empty */
	if (!min_dl_not_nec_local || SMT) {
		if (cs_index >= cs_size) {
			cs_size *= 2;
			cs = (struct constraint*) realloc(cs, cs_size * sizeof(struct constraint));
			assert(cs != NULL);
		}
		cs[cs_index].lhs = (int*) malloc ((set_of_d_paths_index+num_of_resources) * sizeof(int));
		cs[cs_index].pos = (int*) malloc ((set_of_d_paths_index+num_of_resources) * sizeof(int));
		cs[cs_index].lhs_size = 0;
		add_int_to_garbage_bool_int(cs[cs_index].lhs, &garbage);
		add_int_to_garbage_bool_int(cs[cs_index].pos, &garbage);
		bool* participating_resources = (bool*) calloc(num_of_resources, sizeof(bool));
		for (path=0;path<set_of_d_paths_index;path++)
			for(i=0;i<set_of_d_paths[path].num_elts;i++)
				participating_resources[set_of_d_paths[path].elts[i]] = true;

		for (r=0; r<num_of_resources; r++) {
        	if (participating_resources[r]) {
				cs[cs_index].lhs[cs[cs_index].lhs_size] = set_of_d_paths_index+r;//path
				cs[cs_index].pos[cs[cs_index].lhs_size++] = 1;
			}
		}
		cs[cs_index].constr_type = GE;
		cs[cs_index++].rhs = 1;	
		free(participating_resources);
	}
	else {
		#ifdef ZCHAFF
		int clause[set_of_d_paths_index];
		for (path=0;path<set_of_d_paths_index;path++) {
			clause[path] = (path+1)*2+0;
			if (VERBOSE) printf("%i ", path+1);
		}
		cs_index++;
		if (VERBOSE) printf("0\nwhere NUM_OF_LINES = %i\n", cs_index);
		SAT_AddClause(sat, clause, set_of_d_paths_index, 0);
		#endif
	}

	/* Print constraints */
	if (VERBOSE && !min_dl_not_nec_local) {
		printf("Minimize:\n");
			print_constraint(objective);
		printf("Subject to:\n");
		for (i=0;i<cs_index;i++)
			print_constraint(cs[i]);
		printf("where all variables have boolean values.\n");
	}
	else if (SMT) {
		printf("Number of variables = %i, number of constraints = %i\n", set_of_d_paths_index, cs_index);
	}

	/* Interact with solver */
	int ret;
	if (!min_dl_not_nec_local) {
		//LP
		void* lp = init_lp(set_of_d_paths_index+num_of_resources);//TODO strakkere bound
		add_constraint_lp(lp, objective);
		for (i=0;i<cs_index;i++)
			add_constraint_lp(lp, cs[i]);

		if (VERBOSE) printf("Calling solver...\n");
		ret = is_feasible_lp(lp);
		if (ret == RET_DL_MIN && local_minimum)
			ret = RET_DL_NOT_MIN;

		if (ret == RET_DL_NOT_MIN || ret == RET_DL_MIN) {
			printf("%s deadlock found:\n", local_minimum ? "Locally minimal" : ret == RET_DL_MIN ? "Minimal" : "Non-minimal");
			for (path=0;path<set_of_d_paths_index;path++) {
				int value_of_var = get_value_of_var_lp(lp, path);
				if (value_of_var > 0)
					print_worm(set_of_d_paths[path], d_paths_dests[path], params, faulty_resources);
			}
		}
		free_lp(lp);
	}
/*
	else if (SMT) {
		//SMT
		void* smt = init_smt(set_of_d_paths_index);
		for (i=0;i<cs_index;i++)
			add_constraint_smt(smt, cs[i]);
		if (VERBOSE) printf("Calling solver...\n");
		ret = is_feasible_smt(smt);

		if (ret == RET_DL_NOT_MIN || ret == RET_DL_MIN) {
			printf("Non-minimal deadlock found:\n");
			for (path=0;path<set_of_d_paths_index;path++) {
				int value_of_var = get_value_of_var_smt(smt, path);
				if (value_of_var > 0)
					print_worm(set_of_d_paths[path], d_paths_dests[path], params, faulty_resources);
			}
		}
		free_smt(smt);
	}
*/
	else {
		//SAT
		#ifdef ZCHAFF
		int sat_ret = SAT_Solve(sat);
		ret = RET_ERROR;
		switch (sat_ret) {
			case SATISFIABLE:
				ret = RET_DL_NOT_MIN;
				break;
			case UNSATISFIABLE:
				ret = RET_DLF;
				break;
			case UNDETERMINED:
				printf("zchaff could not determine satisfiability of problem.\n");
				break;
			case TIME_OUT:
				printf("zchaff has timed out while determining satisfiability of problem.\n");
				break;
			case MEM_OUT:
				printf("zchaff has timed out while determining satisfiability of problem.\n");
				break;
			case ABORTED:
				printf("zchaff has been aborted while determining satisfiability of problem.\n");
				break;
		}
		int new_set_of_d_paths_size = 0;
		if (ret == RET_DL_NOT_MIN)  {
			printf("Non-minimal deadlock found:\n");
			for (path=0;path<set_of_d_paths_index;path++) {
				int value_of_var = SAT_GetVarAsgnment(sat, path+1);
				if (value_of_var > 0) {
					print_worm(set_of_d_paths[path], d_paths_dests[path], params, faulty_resources);
					new_set_of_d_paths_size++;
				}
			}
		}

		/* Search for local minimum */
		if (LOCAL_MINIMA && ret == RET_DL_NOT_MIN)  {
			if (VERBOSE)
				printf("Searching for local minimum...\n");
			dyn_int_List* new_set_of_d_paths = (dyn_int_List*) malloc (new_set_of_d_paths_size * sizeof(dyn_int_List));
			int * new_d_paths_dests = (int*) malloc (new_set_of_d_paths_size * sizeof(int));
			int new_set_of_d_paths_index = 0;	
			for (path=0;path<set_of_d_paths_index;path++) {
				int value_of_var = SAT_GetVarAsgnment(sat, path+1);
				if (value_of_var > 0) {
					new_set_of_d_paths[new_set_of_d_paths_index] = set_of_d_paths[path];
					new_d_paths_dests[new_set_of_d_paths_index] = d_paths_dests[path];
					new_set_of_d_paths_index++;
				}
			}
			SAT_ReleaseManager(sat);
			ret = interaction_with_solver(false, true, new_set_of_d_paths, new_d_paths_dests, new_set_of_d_paths_index, g, marks, faulty_resources, params);
			free(new_set_of_d_paths);
			free(new_d_paths_dests);
		}
		else
			SAT_ReleaseManager(sat);
		#endif
	}


	/* Free everything */
	free_garbage_bool_int(&garbage);
	if (!min_dl_not_nec_local) {
		free(objective.lhs);
		free(objective.pos);
	}
	free(cs);
	return ret;
}
			


int check_reachability_possible_deadlock (Graph* g, Marks* marks, const Params* params, ResourceList* faulty_resources) {
	int i, j, d, r, num_of_resources = bitsize_of_resources(params);
    bool* visited = new bool[num_of_resources];
	int ret;
	int set_of_d_paths_size = 500;
	dyn_int_List* set_of_d_paths = (dyn_int_List*) malloc (set_of_d_paths_size * sizeof(dyn_int_List));
	int * d_paths_dests = (int*) malloc (set_of_d_paths_size * sizeof(int));
	int set_of_d_paths_index = 0;
	int dests[num_of_resources];
	int num_of_dests;

	for (r=0; r<num_of_resources; r++) {
        if ((marks[r].marking == 4 || marks[r].marking == 3) && inst_convert_to_resource(r,params).p != L) {
			/* For all non-local 3/4 marked channels, enumerate all paths leading to 3 marked channels for all destinations that are deadlock sensitive.*/
			num_of_dests = 0;
			int set_of_d_paths_offset = set_of_d_paths_index;
			for (d=0; d<(int) (marks[r].sens.size()); d++) {
				dyn_int_List *  ret;
				if (!member(marks[r].sens[d],dests,num_of_dests)) {
					dests[num_of_dests++] = marks[r].sens[d];
					for (j=0;j<num_of_resources;j++)
						visited[j] = false;
					ret = enum_paths(r, marks[r].sens[d], g, marks, visited, params, faulty_resources);
				}
				else
					ret = NULL;

				j = 0;
				while (ret != NULL && ret[j].elts != NULL) {
					if (VERBOSE) {
						printf("Path %i: [", set_of_d_paths_index);
						for (i=0;i<ret[j].num_elts;i++)
							std::cout << inst_convert_to_resource(ret[j].elts[i],params) << (i<ret[j].num_elts-1 ? ", " : "");
						printf("] d := ");
						print_procnode(inst_convert_to_procnode(marks[r].sens[d],params),params);
						printf("\n");
					}
					if (set_of_d_paths_index >= set_of_d_paths_size) {
						set_of_d_paths_size *= 2;
						set_of_d_paths = (dyn_int_List*) realloc(set_of_d_paths, set_of_d_paths_size * sizeof(dyn_int_List));
						d_paths_dests = (int*) realloc(d_paths_dests, set_of_d_paths_size * sizeof(int));
						assert(set_of_d_paths != NULL);
						assert(d_paths_dests != NULL);
					}
					d_paths_dests[set_of_d_paths_index]    = marks[r].sens[d];
					set_of_d_paths[set_of_d_paths_index++] = ret[j];
					j++;
				}
				free(ret);
           	}
			/* We have enumerated all paths starting in resource r. We now remove equivalent paths.*/
			remove_equivalent_paths(set_of_d_paths, d_paths_dests, set_of_d_paths_offset, &set_of_d_paths_index, params, faulty_resources);
		}
	}
printf("Num_of_paths = %i\n", set_of_d_paths_index);
//remove_small_paths(set_of_d_paths, d_paths_dests, &set_of_d_paths_index, params, faulty_resources);
//printf("Num_of_paths = %i\n", set_of_d_paths_index);
	delete[] visited;

	/* Call solver to find a legal and reachable deadlock */
	ret = interaction_with_solver(MIN_DL_NOT_NEC, false, set_of_d_paths, d_paths_dests, set_of_d_paths_index, g, marks, faulty_resources, params);


	/* Free everything */
	for (i=0;i<set_of_d_paths_index;i++)
		free(set_of_d_paths[i].elts);
	free(set_of_d_paths);
	free(d_paths_dests);

	return ret;
}
