#include <assert.h>
#include <limits.h>
#include <sys/types.h>
#include <signal.h>

/* 2D mesh topology */
//#include "2d_mesh_xy.c"
//#include "2d_mesh_xy_sp.c"
//#include "2d_mesh_westfirst.c"

/* 2D mesh topology with double channels */
//#include "2dd_mesh_xy_sp.c"

/* 3D mesh topology */
//#include "3d_mesh_deadlock.c"
//#include "3d_mesh_xyz.c"
//#include "3d_mesh_limited_tsp.c"

/* 3D mesh topology with VCs */
//#include "3dvc_mesh_routing.c"

/* NePA Topology */
//#include "abdulaziz_routing4.c"

/* NePA Topology with diagonal links*/
//#include "diagonal_definitions.h"
//#include "diagonal_routing.c"
//#include "diagonal_routing_temp.c"
//#include "diagonal_routing_ft.c"
//#include "diagonal_routing_ft2.c"

/* Spidergon */
//#include "spidergon_sp.c"

/* WNoC */
//#include "wireless_definitions.h"
//#include "wireless_routing.c"
//#include "wireless_routing2.c"
//#include "wireless_vc_routing.c"
//#include "wireless_vc_routing_ft.c"

/* Torus */
//#include "3d_torus_vc_routing_xy.c"
//#include "2d_torus_vc_routing.c"
//#include "temp.c"
//#include "2d_torus_vc_routing_clue.c"/*Only for square meshes*/


/* Duato CE */
//#include "duato_ce_routing.c"

#include "general.h"


//visited[i][d] == -1 means procnode i is unvisited for destination d
//visited[i][d] == k  means procnode i is visited by a message destined for d in run k
int** visited;
int k = 0;
int mentioned_livelock = 0;
int MAX_LIVELOCKS = 1;

extern bool REPORT180;
extern bool ASSEMBLE_STATISTICS;
extern const int RET_DLF;
extern const int RET_DL_MIN;
extern const int RET_DL_NOT_MIN;
extern const int RET_ERROR;
extern const int RET_DISC_ROUTING;
extern const int RET_OK;

/* Adds a label to an edge in the dependency graph structure. Returns 1 iff the label is new, i.e., if it is really added.
   If the edge has not yet been added to the graph, it will automatically be added.

   Inputs:
	g		= the dependency graph structure
	nodeset = an array of bools stating whether node i is an actual node in the graph
    current	= the source of the edge
	neighbor = the end of the edge
	dest = the label to be added
	params = the assisting instantiation specific parameters
*/
int addLabelToEdge(Graph *g, int* nodeset, const Resource &current, const Resource &neighbor, Procnode *dest, const Params* params) {
	int c = inst_convert_resource_to_int(current, params);
	int nbr = inst_convert_resource_to_int(neighbor, params);
	int label = inst_convert_procnode_to_int(dest, params);
	int added = 0, found_nbr = 0, i;


	if (g[c].n == 0) {
		g[c].parents_elements = 0;
		g[c].parents_allocated = 0;
		g[c].parents = NULL;
		nodeset[c] = 1;
	}
	for (i=0;i<g[c].n && added==0;i++) {
		if (g[c].neighbor(i) == nbr) {
			found_nbr = 1;
			assert(g[c].labelCount(i) >= 0 && g[c].labelCount(i) < NUM_OF_PROC_NODES);
			g[c].labels(i)[g[c].labelCount(i)] = label;
            g[c].setLabelCount(i, g[c].labelCount(i)+1);
			added = 1;
		}
	}
	if (found_nbr == 0) {
		int *labels = new int[NUM_OF_PROC_NODES];
		labels[0] = label;
        assert(g[c].n < NUM_OF_OUTGOING_EDGES);
        g[c].set(g[c].n, 0, nbr, 1, labels);
		g[c].n++;
		added = 1;
	}
	//if (added) {
	//	std::cout << "Adding dependency from " << current << " to " << neighbor << " labelled ";
	//	print_procnode(*dest,params);puts("");
	//}
	return added;
}

/* Checks whether `curr' is already in the stac */
int curr_in_stack(int curr, int* stack, int stack_size) {
	int i,ret=0;
	for (i=0;i<stack_size&&ret==0;i++)
		ret = stack[i] == curr ? 1 : 0; 
	return ret;
}


/* Simulates the injection of a packet currently in `curr'.
   At each recursive call, the next hops are computed by the routing function.
   We check:
	1.) if al the next hops are valid resources
	2.) if each hop is correct wrt to the topology, i.e., message do not fly from resource to resource
	3.) if the routing function always knows at least one next hop
*/
int simulate_injection(const Resource &curr, Procnode *dest, Graph *g, int* nodeset, ResourceList* faulty_resources, const Params* params, int* stack, int stack_size) {
	Procnode curr_end = get_end(curr);
	int i = 0, newLabel, num_of_neighbors = 0, dest_int = inst_convert_procnode_to_int(dest, params), curr_int = inst_convert_resource_to_int(curr, params);
	int ret = RET_OK;

if (stack_size > 0) {
	g[curr_int].parents_elements = 0;
	g[curr_int].parents_allocated = 0;
	g[curr_int].parents = NULL;
	nodeset[curr_int] = 1;
}

	// If the packet gets into a cycle, we have a livelock
	// Print the stack to output the livelock
	if (mentioned_livelock < MAX_LIVELOCKS && visited[curr_int][dest_int] == k && curr_in_stack(curr_int,stack,stack_size) == 1) {
		mentioned_livelock++;
		printf("Livelock detected for destination ");
		print_procnode(*dest,params);
		printf(":\n");
		for (i=0;i<stack_size;i++) {
            std::cout << inst_convert_to_resource(stack[i],params);
			printf(", ");
		}
        std::cout << curr;
		if (mentioned_livelock >= MAX_LIVELOCKS)
			printf("\nSuppressing detection of other live-locks.\n");
		else
			printf("\n");
	}
	// If simulations for the current resource/destination has already been done, than:
	// 1.) there is no further livelock possible (it would already have been mentioned)
	// 2.) all possible dependencies yeilded by these simulations ahve already been added
	// So we can return;
	if (visited[curr_int][dest_int] != -1) 
		return ret;
	// Mark the current resource as visited in the current run of the simulations
	visited[curr_int][dest_int] = k;
	// Add it to the stack for livelock detection
	stack[stack_size] = curr_int;
    
	ResourceList *neighbors = inst_routing(curr,dest,params,faulty_resources);



	// For each next hop, perform the checks and prcoeed with simulation
	while (ret == RET_OK && neighbors->at(i).valid()) {
 		if (is_faulty_resource(neighbors->at(i), faulty_resources) != 1) {
			newLabel = addLabelToEdge(g,nodeset,curr,neighbors->at(i),dest,params);
			num_of_neighbors++;
			if (newLabel == 1) {
				// Check if the nexthop is valid:
				if (valid_resource(neighbors->at(i))!= 1) {
					printf("The routing function generates invalid resources.\n");
					printf("A packet with destination ");
					print_procnode(*dest, params);
					printf(" routes from ");
            	    std::cout << curr;
					printf(" to illegal neighbor:\n");
					std::cout << neighbors->at(i);
					printf("\nThis neighbor is not valid.\n");
					kill(0,SIGTERM);
				}
				// Check if the nexthop actually respects the topology
				Procnode next_src = get_src(neighbors->at(i));
				if (equal_procnodes(curr_end, next_src) != 1) {
					printf("The routing function is not correct with respect to the topology.\n");
					printf("A packet with destination ");
					print_procnode(*dest, params);
					printf(" routes from ");
					std::cout << curr;
					printf(" to ");
					std::cout << neighbors->at(i);
					printf(".\n");
					kill(0,SIGTERM);
				}
				// Check if a 180 turn has been made 
				if (REPORT180 && equal_procnodes(get_src(curr), get_end(neighbors->at(i)))) {
					printf("Made a 180 turn.\n");
					printf("A packet with destination ");
					print_procnode(*dest, params);
					printf(" routes from ");
					std::cout << curr;
					printf(" to ");
					std::cout << neighbors->at(i);
					printf(".\n");
					kill(0,SIGTERM);
				}
				ret = simulate_injection(neighbors->at(i),dest,g,nodeset,faulty_resources,params,stack,stack_size+1);
			}
		}
		i++;
	}

	// Check if there is always one nexthop if we have not arrived at the destination
	if (ret == RET_OK && num_of_neighbors==0 && equal_procnodes(curr_end, *dest) != 1) {
        if (output) {
			std::cout << "Disconnected routing function!" << std::endl;
    	    std::cout << "No route from " << curr << " to ";
			print_procnode(*dest, params);
        	std::cout << " with faulty resources " << *faulty_resources << std::endl;
        	std::cout << "The set of possible next hops is: " << *inst_routing(curr,dest,params,faulty_resources) << std::endl;
		}
		ret = RET_DISC_ROUTING;
	}
	
	delete neighbors;
	return ret;
}

int fill_graph(Graph *g, int*nodeset, ResourceList* faulty_resources, const Params* params) {
	int i,j, ret = RET_OK;
    int nr = bitsize_of_resources(params);
    int np = bitsize_of_procnodes(params);
    int* stack = new int[np];
	visited = new int*[nr];
	for (i=0;i<nr;i++) {
		nodeset[i] = 0;
		g[i].n = 0;
		int *nw = new int[np];
        //memset_pattern4(visited[i], -1, np);
		for (j=0;j<np;j++)
			nw[j] = -1;
        visited[i] = nw;
	}



	Procnode ** p = inst_get_procnodes(params);
	for (i=0;i<inst_num_of_procnodes(params);i++) {
		if (p[i] != NULL) {
			Resource r = inst_add_local_to_procnode(p[i]);
			for (j=0;j<inst_num_of_procnodes(params);j++) {
				if (p[j] != NULL) {
					// Note that we do not have to clean the stack (set it to -1). This is important for efficiency, saves a degree.
					//printf("Simulations for node ");print_procnode(*(p[i]),params);puts("");
					ret = simulate_injection(r, p[j], g, nodeset, faulty_resources, params, stack, 0);
					if (ret != RET_OK)
						goto TERMINATE_FILL_GRAPH;
					assert(k < INT_MAX);
					k += 1;
				}
			}
		}
	}
TERMINATE_FILL_GRAPH:
	mentioned_livelock = 0;
	k = 0;
	// Free everything
	for (i=0;i<inst_num_of_procnodes(params);i++)
        if (p[i])
            delete p[i];
	delete[] p;
	for (i=0;i<nr;i++)
		delete visited[i];
	delete[] visited;
	delete[] stack;
	return ret;
}


void free_graph(Graph* g, int* nodeset, const Params* params) {
	int i,j;
	for (i=0;i<bitsize_of_resources(params);i++) {
		if (nodeset[i] == 1) {
			for(j=0;j<g[i].n;j++) {
				free(g[i].labels(j));
			}
			//free(g[i].edges);
			if (g[i].parents_allocated > 0) {
				free(g[i].parents);
			}
		}
	}
}



// Unused function, but handy for debugging
void valid_graph(int num_of_resources, Graph *g, int *nodeset, Marks *marks) {
	int i,j;
	for (i=0;i<num_of_resources;i++) {
		for(j=0;j<g[i].n;j++) {
			if (nodeset[g[i].neighbor(j)] == 1 && !(g[i].neighbor(j) >= 0 && g[i].neighbor(j) < num_of_resources)) {
				fprintf(stderr,"ERROR: neighbor out of bound (%i,%i)\n", g[i].neighbor(j), num_of_resources);
			}
			if (g[i].neighbor(j) == i)
				fprintf(stderr,"ERROR: neighbor equal to node (%i)\n", i);
		}
		if (nodeset[i] != 0 && nodeset[i] != 1)
			fprintf(stderr,"ERROR: nodeset not initialized (%i)\n", i);
		if (marks[i].marking != 0 && marks[i].marking != 1 && marks[i].marking != 2 && marks[i].marking != 3 && marks[i].marking != 4)
			fprintf(stderr,"ERROR:\n");
	}
}




/* Add the resource to the parent-array of the current resource
Dynamically resizes this array if necessary  */
void addToParents(int resource, int parent, Graph *g) {
	if(g[resource].parents_elements == g[resource].parents_allocated) {	// Are more refs required?
		if (g[resource].parents_allocated == 0) {						// Initialize list
			g[resource].parents_allocated = 4; 
		}
		else {															// Expand list, each time at least doubling the size
			g[resource].parents_allocated *= 2;
		}
		
		void *_tmp = realloc(g[resource].parents, (g[resource].parents_allocated * sizeof(int)));
	    if (!_tmp)
		{ 
			fprintf(stderr, "ERROR: Couldn't realloc memory!\n");puts("");
			return; 
		}
		g[resource].parents = (int*)_tmp;	
	}

	g[resource].parents[g[resource].parents_elements] = parent;
	g[resource].parents_elements++;
}



bool contains_neighbor(int parent, int neighbor, const Graph *g) {
    const Graph &current = g[parent];
    int n = current.n;
	for (int i=n-1; i >= 0; --i) {
		if (current.neighbor(i) == neighbor)
			return true;
	}
	return false;
}

/* Updates the graph struct with parents. */
void compute_parents (Graph *g, int *nodeset, int num_of_resources) {
	int i,j;
	for (i=0;i<num_of_resources;i++) {
		if (nodeset[i] == 1) {
			for (j=0;j<num_of_resources;j++) {
				if (nodeset[j] == 1) {
                    if (contains_neighbor(j, i, g)) {
						addToParents(i, j, g);
					}
				}
			}
		}
	}
}
