
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <assert.h>
#include "ilcplex/cplex.h"
#include "datastructures.h"


//extern bool MIN_DL_NOT_NEC;
extern const int RET_DLF;
extern const int RET_DL_MIN;
extern const int RET_DL_NOT_MIN;
extern const int RET_ERROR;

volatile int terminator = 0;
struct cplex_lp {
	double* obj;
	double* lb;
	double* ub;
    int *rmatbeg;
    int *rmatind;
    double *rmatval;
    double *rhs;
    char *sense;
	char* ctype;

	int num_of_cols;
	int num_of_rows;
	int num_of_rows_allocated;
	int num_nz;
	int num_nz_allocated;

	double* solution;
};

void* init_lp(int num_of_vars) {
	struct cplex_lp* lp = (struct cplex_lp*) malloc(sizeof(struct cplex_lp));
	
	int INIT_NUM_ROWS = 1000;
	int INIT_NZ= 10000;

	lp->obj		 = (double*) malloc (num_of_vars * sizeof(double));
	lp->lb		 = (double*) malloc (num_of_vars * sizeof(double));
	lp->ub		 = (double*) malloc (num_of_vars * sizeof(double));
	lp->rmatbeg  = (int*) malloc (INIT_NUM_ROWS * sizeof(int));
   	lp->rmatind  = (int*) malloc (INIT_NZ * sizeof(int));
   	lp->rmatval  = (double*) malloc (INIT_NZ * sizeof(double));
    lp->rhs		 = (double*) malloc (INIT_NUM_ROWS * sizeof(double));
    lp->sense	 = (char*) malloc (INIT_NUM_ROWS * sizeof(char));
	lp->ctype    = (char*) malloc (num_of_vars * sizeof(char));
	lp->solution = NULL;
	
	lp->num_of_cols = num_of_vars;
	lp->num_of_rows = 0;
	lp->num_of_rows_allocated = INIT_NUM_ROWS;
	lp->num_nz = 0;
	lp->num_nz_allocated = INIT_NZ;

	lp->solution = NULL;

	/* Set the lower- and upper bounds of all columns. Also, set them to binary */
	for (int i=0;i<num_of_vars;i++) {
		lp->lb[i] = 0.0;
		lp->ub[i] = 1.0;
		lp->ctype[i] = 'B';
	}

	return (void*) lp;
}



static int CPXPUBLIC mycallback (CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle) {
	int c;
	if ((c=getchar()) != EOF) {
		switch (c) {
			case 'q':
				terminator = 1;
				break;
			case 'w':
				int feasible;
				int status = CPXgetcallbackinfo (env, cbdata, wherefrom, CPX_CALLBACK_INFO_MIP_FEAS, &feasible);
 				assert(!status);
				if (feasible)
					printf("A legal and reachable deadlock has been found, not necessarily minimal.\n");
				else
					printf("No deadlock has been found yet.\n");
				break;
		}
	}
	return 0;
}

void add_constraint_lp(void*lp_pointer, struct constraint c) {
	struct cplex_lp* lp = (struct cplex_lp*) lp_pointer;

	while (lp->num_of_rows >= lp->num_of_rows_allocated) {
		lp->num_of_rows_allocated *= 2;
		lp->rmatbeg  = (int*) realloc (lp->rmatbeg, lp->num_of_rows_allocated  * sizeof(int));
		lp->rhs      = (double*) realloc (lp->rhs, lp->num_of_rows_allocated  * sizeof(double));
		lp->sense    = (char*) realloc (lp->sense, lp->num_of_rows_allocated  * sizeof(char));
		assert (lp->rmatbeg != NULL);
		assert (lp->rhs != NULL);
		assert (lp->sense != NULL);
	}
	while (lp->num_nz + lp->num_of_cols >= lp->num_nz_allocated) {
		lp->num_nz_allocated *= 2;
		lp->rmatind   = (int*) realloc (lp->rmatind, lp->num_nz_allocated  * sizeof(int));
		lp->rmatval   = (double*) realloc (lp->rmatval, lp->num_nz_allocated  * sizeof(double));
		assert (lp->rmatind != NULL);
		assert (lp->rmatval != NULL);
	}

  	int i;
	if (c.constr_type == OBJ) {
		assert(lp->num_of_cols>0);
		for (i=0;i<lp->num_of_cols;i++)
			lp->obj[i] = 0;
		assert (c.lhs_size>0);
		for (i=0;i<c.lhs_size;i++) {
			assert(c.lhs[i]<lp->num_of_cols);
			if (c.lhs[i] >= 0)
				lp->obj[c.lhs[i]] = c.pos[i];
		}
	}
	else {
		int rhs = c.rhs;
		/* The start of the current row */
		assert (c.lhs_size>0);
		lp->rmatbeg[lp->num_of_rows] = lp->num_nz;
		bool check = false;
		for (i=0;i<c.lhs_size;i++) {
			if (c.lhs[i] >= 0) {
				check = true;
				assert(c.lhs[i]<lp->num_of_cols);
				lp->rmatind[lp->num_nz] = c.lhs[i];
				lp->rmatval[lp->num_nz] = c.pos[i];
				lp->num_nz++;
			}
			else
				/* Subtract the value from the rhs */
				rhs -= c.pos[i] * -(c.lhs[i] + 1);
		}
		assert(check);
		/* The rhs of the constraint */
		lp->rhs[lp->num_of_rows] = rhs;
		lp->sense[lp->num_of_rows] = c.constr_type == LE ? 'L' : c.constr_type == GE ? 'G' : 'E';
		lp->num_of_rows++;
	}
}


void print_lp(void* lp) {
	//TODO
}



int is_feasible_lp(void* lp_pointer) {
	struct cplex_lp* lp = (struct cplex_lp*) lp_pointer;
	
	CPXENVptr     env = NULL;
	CPXLPptr      lp_obj = NULL;
	int           status = 0;

	/* Initialize the CPLEX environment */
	env = CPXopenCPLEX (&status);

   	/* If an error occurs, the status value indicates the reason for
      failure.  A call to CPXgeterrorstring will produce the text of
      the error message.  Note that CPXopenCPLEX produces no output,
      so the only way to see the cause of the error is to use
      CPXgeterrorstring.  For other CPLEX routines, the errors will
      be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */
   	if ( env == NULL ) {
    	char  errmsg[CPXMESSAGEBUFSIZE];
    	fprintf (stderr, "Could not open CPLEX environment.\n");
    	CPXgeterrorstring (env, status, errmsg);
    	fprintf (stderr, "%s", errmsg);
		assert(false);
	}

   	/* Turn on data checking *///TODO
   	status = CPXsetintparam (env, CPX_PARAM_DATACHECK, CPX_ON);
   	if (status)
    	fprintf (stderr, "Failure to turn on data checking, error %d.\n", status);

 	/* Turn on output to the screen */
   	status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
   	if (status) {
    	fprintf (stderr, "Failure to turn on screen indicator, error %d.\n", status);
		assert(false);
	}

	/* Create the problem. */
	lp_obj = CPXcreateprob (env, &status, "deadlock");

   	/* A returned pointer of NULL may mean that not enough memory
      was available or there was some other problem.  In the case of 
      failure, an error message will have been written to the error 
      channel from inside CPLEX.  In this example, the setting of
      the parameter CPX_PARAM_SCRIND causes the error message to
      appear on stdout.  */
   	if (lp_obj == NULL) {
    	fprintf (stderr, "Failed to create LP.\n");
		assert(false);
   	}

	/* Problem is minimization */
   	CPXchgobjsen (env, lp_obj, CPX_MIN);  

//CPXsetintparam(env, CPX_PARAM_MIPEMPHASIS, CPX_MIPEMPHASIS_FEASIBILITY);
//CPXsetdblparam(env, CPX_PARAM_OBJLLIM, 7.0);
//CPXsetdblparam(env, CPX_PARAM_EPAGAP, 2.0);
//assert(!status);


   	/* Now create the new columns.  First, populate the arrays. */
   	status = CPXnewcols(env, lp_obj, lp->num_of_cols, lp->obj, lp->lb, lp->ub, NULL, NULL);
    if (status) {
    	fprintf (stderr, "Failed to create columns.\n");
		assert(false);
	}

   	/* Now populate the problem with the data.  For building large
      problems, consider setting the row, column and nonzero growth
      parameters before performing this task. */
	status = CPXaddrows(env, lp_obj, 0, lp->num_of_rows, lp->num_nz, lp->rhs, lp->sense, lp->rmatbeg, lp->rmatind, lp->rmatval, NULL, NULL);
    if (status) {
    	fprintf (stderr, "Failed to populate problem.\n");
		assert(false);
	}
   
	/* Now copy the ctype array */
   	status = CPXcopyctype (env, lp_obj, lp->ctype);
   	if ( status ) {
    	fprintf (stderr, "Failed to copy ctype\n");
		assert(false);
   	}

//status = CPXwriteprob (env, lp_obj, "temp.mps", NULL);

	/* We add a callback and set the terminal in order to be able to quit the solver by pressing q */
   	status = CPXsetinfocallbackfunc(env, mycallback, NULL);
	status = CPXsetterminate (env, &terminator);
   	if (status) {
    	fprintf (stderr, "Failed to set callback function.\n");
		assert(false);
   	}
	// Save the existing fcntl flags
	int saved_flags = fcntl(0, F_GETFL);
	// Set the new flags with the O_NONBLOCK bit set
	fcntl(0, F_SETFL, saved_flags | O_NONBLOCK);	 
	struct termios oldt;
	struct termios newt;
	/* Save the existing terminal setting */
	tcgetattr(STDIN_FILENO, &oldt); 
	/* Copy old settings to new settings */
	newt = oldt;
	/* Make one change to old settings in new settings: enable canonical mode to support reading of a single character */
	newt.c_lflag &= ~ICANON;
	/* Apply the new settings immediatly */
	tcsetattr(STDIN_FILENO, TCSANOW, &newt);


   	/* Optimize the problem and obtain solution. */
   	status = CPXmipopt (env, lp_obj);
   	if (status) {
    	fprintf (stderr, "Failed to optimize MIP.\n");
		assert(false);
   	}

	/* Reapply the old settings */
	fcntl(0, F_SETFL, saved_flags);	 
	tcsetattr(STDIN_FILENO, TCSANOW, &oldt);



   	/* Analyze and store the output.
      The size of the problem should be obtained by asking CPLEX what
      the actual size is, rather than using what was passed to CPXcopylp.
      cur_numrows and cur_numcols store the current number of rows and
      columns, respectively.  */
	int solstat = CPXgetstat(env, lp_obj);
	int ret = solstat == CPXMIP_OPTIMAL ||
			solstat == CPXMIP_OPTIMAL_TOL ||
			solstat == CPXMIP_FEASIBLE ||
			solstat ==  CPXMIP_NODE_LIM_FEAS ||
			solstat == 	CPXMIP_TIME_LIM_FEAS ||
			solstat == CPXMIP_FAIL_FEAS ||
			solstat == CPXMIP_MEM_LIM_FEAS ||
			solstat ==  CPXMIP_ABORT_FEAS ||
			solstat ==  CPXMIP_FAIL_FEAS_NO_TREE;

   	if (ret) {
		int cur_numcols = CPXgetnumcols (env, lp_obj);
assert(cur_numcols == lp->num_of_cols);
		double* x = (double*) malloc(cur_numcols * sizeof (double));
		status = CPXgetx (env, lp_obj, x, 0, cur_numcols-1);
	   	if (status)
			return 0;
		lp->solution = x;
   		
		double objval;
	   	status = CPXgetobjval (env, lp_obj, &objval);
   		if (status) {
			fprintf (stderr,"No MIP objective value available.  Exiting...\n");
			assert(false);
   		}
	}
	if (terminator != 0 && !ret)
		printf("Terminated CPLEX and no solution has yet been found. The deadlock may be a possible legal reachable deadlock.\n");
	if (terminator != 0 && ret)
		printf("Terminated CPLEX and a solution has yet been found. The deadlock is legal and reachable, but not necessarily minimal.\n");


  	/* Free up the problem as allocated by CPXcreateprob, if necessary */
   	if (lp_obj != NULL) {
    	status = CPXfreeprob(env, &lp_obj);
    	if (status)
        	fprintf (stderr, "CPXfreeprob failed, error code %d.\n", status);
   	}
   	/* Free up the CPLEX environment, if necessary */
   	if (env != NULL) {
    	status = CPXcloseCPLEX (&env);
    	/* Note that CPXcloseCPLEX produces no output,
         so the only way to see the cause of the error is to use
         CPXgeterrorstring.  For other CPLEX routines, the errors will
         be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON. */
      	if (status) {
        	char errmsg[CPXMESSAGEBUFSIZE];
        	fprintf (stderr, "Could not close CPLEX environment.\n");
        	CPXgeterrorstring (env, status, errmsg);
        	fprintf (stderr, "%s", errmsg);
    	}
   	}

	return terminator != 0 ? (ret ? RET_DL_NOT_MIN : RET_ERROR) : (ret ? RET_DL_MIN : RET_DLF); 
}



void free_lp(void* lp_pointer) {
	struct cplex_lp* lp = (struct cplex_lp*) lp_pointer;
 	free(lp->obj);
	free(lp->lb);
	free(lp->ub);
	free(lp->rmatbeg);
   	free(lp->rmatind);
   	free(lp->rmatval);
    free(lp->rhs);
    free(lp->sense);
	free(lp->ctype);
	free(lp->solution);
	free(lp);
}

int get_value_of_var_lp(void* lp_pointer, int var) {
	struct cplex_lp* lp = (struct cplex_lp*) lp_pointer;
	if (lp->solution != NULL)
		return lp->solution[var];
	return -1;
}
