#include "lp_solve/lp_lib.h"

//extern bool MIN_DL_NOT_NEC;
extern const int RET_DLF = 0;
extern const int RET_DL_MIN = 1;
extern const int RET_DL_NOT_MIN = 2;
extern const int RET_ERROR = -1;

int* index_of_vars;


void* init_lp(int num_of_vars) {
	lprec* lp = make_lp(0,0);
	set_add_rowmode(lp, TRUE);
	set_verbose(lp, IMPORTANT);
	if (MIN_DL_NOT_NEC)
		set_break_at_first(lp, TRUE);
	index_of_vars = (int*) malloc(num_of_vars * sizeof(int));
	int i;
	for (i=0;i<num_of_vars;i++)
		index_of_vars[i] = -1;
	set_bounds_tighter(lp, TRUE);
	return (void*) lp;
}

void add_constraint_lp(void*lp_pointer, struct constraint c) {
	lprec* lp = (lprec*) lp_pointer;
	REAL* constraint = (REAL*) (malloc (c.lhs_size * (sizeof (REAL))));
	int *constraint_indices = (int*) (malloc (c.lhs_size * (sizeof (int))));
	int constraint_size = 0;
	int rhs = c.rhs;
	
	int i;
	for (i=0;i<c.lhs_size;i++) {
		/* The variable name */
		if (c.lhs[i] >= 0) {
			/* Lookup the variable name */
			if (index_of_vars[c.lhs[i]] == -1) {
				char name_of_var[10];
				sprintf(name_of_var, "x%i", c.lhs[i]);
				add_columnex(lp, 0, NULL, NULL);
				index_of_vars[c.lhs[i]] = 1+get_Ncolumns(lp);
				set_col_name(lp,index_of_vars[c.lhs[i]],name_of_var);
				set_binary(lp,index_of_vars[c.lhs[i]],TRUE);
			}

			constraint[constraint_size] = c.pos[i];
			constraint_indices[constraint_size++] = index_of_vars[c.lhs[i]];
		}
		else
			/* Subtract the value from the rhs */
			rhs -= c.pos[i] * -(c.lhs[i] + 1);
	}
	if (c.constr_type == OBJ)
		set_obj_fnex(lp, constraint_size, constraint, constraint_indices);
	else 
		add_constraintex(lp, constraint_size, constraint, constraint_indices, c.constr_type, (REAL) rhs);
	free(constraint_indices);
	free(constraint);
}


void print_lp(void* lp) {
	write_lp((lprec*) lp, NULL);
}

int is_feasible_lp(void* lp_pointer) {
	lprec* lp = (lprec*) lp_pointer;
	set_add_rowmode(lp, FALSE);
	set_presolve(lp, PRESOLVE_ROWS | PRESOLVE_COLS | PRESOLVE_LINDEP | PRESOLVE_SOS, get_presolveloops(lp));
	int ret = solve(lp);
	return ret == OPTIMAL || ret == SUBOPTIMAL || ret == PRESOLVED || ret == PROCBREAK || ret == FEASFOUND ? RET_DL_MIN : RET_DLF;
}

void free_lp(void* lp_pointer) {
	lprec* lp = (lprec*) lp_pointer;
	delete_lp(lp);
}

int get_value_of_var_lp(void* lp_pointer, int var) {
	lprec* lp = (lprec*) lp_pointer;
	char* name_of_var = (char*) malloc(10 *sizeof (char));
	sprintf(name_of_var, "x%i", var);

	REAL* var_values;
	get_ptr_variables(lp, &var_values);
	for (int i=0;i<get_Ncolumns(lp);i++) {
		if (!strcmp(name_of_var, get_col_name(lp, i+1))) {
			free(name_of_var);
			return (int) var_values[i];
		}
	}
	free(name_of_var);
	return -1;
}
