#include "yices/include/yices_c.h"
#include "gen_dep_graph.c"

//extern bool MIN_DL_NOT_NEC;
extern bool VERBOSE;
extern const int RET_DLF;
extern const int RET_DL_MIN;
extern const int RET_DL_NOT_MIN;
extern const int RET_ERROR;

// We cache some variable declarations, this saves many gigabytes of memory
yices_var_decl* var_decl;
yices_type yices_nat_type;
yices_expr yices_zero;

void* init_smt(int num_of_vars) {
	yices_context smt = yices_mk_context();
	var_decl = (yices_var_decl*) malloc (num_of_vars * sizeof(yices_var_decl));
	int i;
	for (i=0;i<num_of_vars;i++)
		var_decl[i] = 0;
	yices_nat_type = yices_mk_type(smt, "nat");//yices_parse_type(smt, "(subtype (n::nat) (<= n 1))"));
	yices_zero = yices_mk_num(smt, 0);
	return smt;
}



void add_constraint_smt(void*smt_pointer, struct constraint c) {
	yices_context smt = smt_pointer;
	yices_expr sum[c.lhs_size];
	int sum_index = 0;
	int rhs = c.rhs;
	
	int i;
	for (i=0;i<c.lhs_size;i++) {
		/* The variable name */
		if (c.lhs[i] >= 0) {
			/* Lookup the variable name */
			yices_expr e;
			if (var_decl[c.lhs[i]] == 0) {
				/* A new variable. Add the var_decl to the cache and assert that it is less than 2. */
				char name_of_var[10];
				sprintf(name_of_var, "x%i", c.lhs[i]);
				var_decl[c.lhs[i]] = yices_mk_var_decl(smt, name_of_var, yices_nat_type);
assert(var_decl[c.lhs[i]] != 0);
				e = yices_mk_var_from_decl(smt, var_decl[c.lhs[i]]);
				yices_assert(smt, yices_mk_le(smt, e, yices_mk_num(smt, 1)));
			}
			e = yices_mk_var_from_decl(smt, var_decl[c.lhs[i]]);
			assert(sum_index<c.lhs_size);
			if (c.pos[i] > 0)
				sum[sum_index++] = e;
			else {
				yices_expr negate[2];
				negate[0] = yices_zero;
				negate[1] = e;
				sum[sum_index++] = yices_mk_sub(smt, negate, 2);
			}
		}
		else
			/* Subtract the value from the rhs */
			rhs -= c.pos[i] * -(c.lhs[i] + 1);
	}
	if (c.constr_type == OBJ)
		assert(false);
	else {
		yices_expr lhs_e = yices_mk_sum(smt, sum, sum_index);
		yices_expr rhs_e = yices_mk_num(smt, rhs);
		yices_expr c_e =	c.constr_type == LE ? yices_mk_le(smt, lhs_e, rhs_e) :
					 		c.constr_type == EQ ? yices_mk_eq(smt, lhs_e, rhs_e) :
							yices_mk_ge(smt, lhs_e, rhs_e);

		if (VERBOSE) {
			printf("Adding yices constraint: ");
			yices_pp_expr(c_e);
			puts("");
		}

		yices_assert(smt, c_e);
	}
}


void print_smt(void* smt) {
}

int is_feasible_smt(void* smt_pointer) {
	yices_context smt = smt_pointer;
	switch(yices_check(smt)) {
		case l_true:
			return RET_DL_NOT_MIN;
		case l_false:
			return RET_DLF;
		case l_undef:
			return RET_ERROR;
	}
	return RET_ERROR;
}

void free_smt(void* smt_pointer) {
	yices_context smt = smt_pointer;
	yices_del_context (smt);
}

int get_value_of_var_smt(void* smt_pointer, int var) {
	yices_context smt = smt_pointer;
	char* name_of_var = (char*) malloc(10 *sizeof (char));
	sprintf(name_of_var, "x%i", var);
	yices_var_decl vd = yices_get_var_decl_from_name(smt, name_of_var);
	if (vd == 0) return -1;
	long ret;
	yices_get_int_value(yices_get_model(smt), vd, &ret);

	free(name_of_var);
	return (int) ret;
}
