/**************/
/* XY ROUTING */
/**************/
#include "2d_mesh_definitions.h"

/* The routing function. Routes from channel c == (x,y,p) to processing node d == (dx,dy)
   Returns a set of resources. As xy routing is deterministic, only one next hop is returned.
   This list of one element is closed with a NULL.
*/
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<2>;
    Resource *nexthops = hops->routed;


	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;

	if (dx != nx || dy != ny) {						// If we do not arrive at our destination
		if (dx < nx)
			nexthops[0] = Resource(nx,ny,W);	//go west if the destination is west
		else if (dx > nx)
			nexthops[0] = Resource(nx,ny,E);	//go east if the destination is east
		else if (dy < ny)
			nexthops[0] = Resource(nx,ny,N);	//go north if the destination is north && in the right column
		else if (dy > ny)
			nexthops[0] = Resource(nx,ny,S);	//go south if the destination is south && in the right column
	}
	return hops;
}
