
/**********************/
/* WEST-FIRST ROUTING */
/**********************/

#include "2d_mesh_definitions.h"



// The routing function.
// Note that routing is not memoryless, i.e., it depends on the port p instead of the current procnode (x,y).
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<5>;
    Resource *nexthops = hops->routed;

	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;
	// For westfirst we also need the port of the current channel, i.e., routing is not memoryless
	int p = c.p;
	int i = 0;

	if (dx != nx || dy != ny) {										// If we do not arrive at our destination
		if (nx > 0 && (p == W || p == L))							// Go west only if we are in western direction or just injected
			nexthops[i++] = Resource(nx,ny,W);
		if (dx > nx)												// Go east if the destination is east of us
			nexthops[i++] = Resource(nx,ny,E);
		if (ny > 0 && p != S && (dx > nx || (dx == nx && dy < ny)))	// Go north if the destination is east or north of us. Prohibit south-north turns.
			nexthops[i++] = Resource(nx,ny,N);
		if (ny < (dim->y - 1) && (dx > nx || (dx == nx && dy > ny)))// Go south if the destination is east or south of us.
			nexthops[i++] = Resource(nx,ny,S);
	}
	return hops;
}


