/*********************/
/* 2D TORUS ROUTING */
/********************/
#include<math.h>
#include "2d_torus_vc_definitions.h"
int ksize=PARAM_x;

double CD(Procnode p){					//Calculate the distance from current node to the center of the networks
	int x=p.x, y=p.y;
	return ((ksize-1)/2.0-x)*((ksize-1)/2.0-x)+((ksize-1)/2.0-y)*((ksize-1)/2.0-y);
}

int checkWraplink(int xoff,int yoff){  	//Check if the message need to pass the wraparound links
	int cnt=0;
	if((abs(xoff)>ksize/2.0)||(abs(yoff)>ksize/2.0))
		cnt=1;
	return cnt;
}

/* The routing function. Routes from channel c == (x,y,p) to processing node d == (dx,dy)
   Returns a set of resources. This list of one element is closed with a NULL.
*/

ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *vc1_hops 	   = new ResourceListN<6>;
    Resource 	 *vc1_nexthops = vc1_hops->routed;
	
    ResourceList *hops 	       = new ResourceListN<6>;
    Resource 	 *nexthops     = hops->routed;
	int max_num_of_nexthops    = 6;
	int vc1_nexthops_flag[6]   = {0,0,0,0,0,0};
	
	
	int i=0,vc1_i=0;

	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;
	// The offset of the desination node  and the processing node
	int xoff=dx-nx, yoff=dy-ny;

	if (xoff != 0 || yoff != 0 ) {										// If we do not arrive at our destination
		if (checkWraplink(xoff, yoff)!= 0){                              // Do need to traverse the wraparound link
			if ((xoff>0&&xoff<=ksize/2.0)||xoff<-ksize/2.0){			
				vc1_nexthops[vc1_i] = Resource(nx,ny,E1);				
				vc1_nexthops_flag[vc1_i++] = 1;
			}
			if ((xoff<0&&xoff>=-ksize/2.0)||xoff>ksize/2.0){
				vc1_nexthops[vc1_i] = Resource(nx,ny,W1);			
				vc1_nexthops_flag[vc1_i++] = 1;
			}
			if ((yoff>0&&yoff<=ksize/2.0)||yoff<-ksize/2.0){
				vc1_nexthops[vc1_i] = Resource(nx,ny,N1);
				vc1_nexthops_flag[vc1_i++] = 1;
			}
			if ((yoff<0&&yoff>=-ksize/2.0)||yoff>ksize/2.0){
				vc1_nexthops[vc1_i] = Resource(nx,ny,S1);	
				vc1_nexthops_flag[vc1_i++] = 1;
			}
			

			// Provide the direction which lead the packet to the right direction for VC1
			int it;
			for(it = 0;it < max_num_of_nexthops;it++){
				if((vc1_nexthops_flag[it] != 0)&&(CD(get_end(vc1_nexthops[it])) >= CD(get_src(vc1_nexthops[it])))){
					nexthops[i++] = vc1_nexthops[it];
				}
			}
			

			// Provide the wraparound linke when the node is the bound node in the lowest dimension(need to pass wraparound links)
			if(xoff > ksize/2.0 && nx == 0){
				nexthops[i++] = Resource(nx,ny,W2);
			}
			if(xoff < -ksize/2.0 && nx == ksize-1){
				nexthops[i++] = Resource(nx,ny,E2);
			}
			if(fabs(xoff) <= ksize/2.0){
				if(yoff > ksize/2.0 && ny == 0){
					nexthops[i++] = Resource(nx,ny,S2);
				}
				if(yoff < -ksize/2.0&& ny == ksize-1){
					nexthops[i++] = Resource(nx,ny,N2);
				}
			}
			
		}
		else{								    			// Do not need to traverse the wraparound link
			if(xoff < 0){
				nexthops[i++] = Resource(nx,ny,W1);
			}
			if(xoff > 0){
				nexthops[i++] = Resource(nx,ny,E1);
			}
			if(yoff < 0){
				nexthops[i++] = Resource(nx,ny,S1);
			}
			if(yoff > 0){
				nexthops[i++] = Resource(nx,ny,N1);
			}
				
			if(xoff != 0){
				if(xoff < 0){
					nexthops[i++] = Resource(nx,ny,W2);
				}
				if(xoff > 0){				
					nexthops[i++] = Resource(nx,ny,E2);
				}
			}
			else{
				if(yoff < 0){
					nexthops[i++] = Resource(nx,ny,S2);
				}
				if(yoff > 0){
					nexthops[i++] = Resource(nx,ny,N2);
				}
			}
		}
	}
	return hops;
}


