/**************/
/* ROUTING    */
/**************/

/*
Routing function for Diagonal Mesh NoC from UCI.
Fault-tolerant up to two faults, uses a look-a-head of 1.
*/

#include "diagonal_faults.c"


int is_border_channel(int x, int y, port p, const Params* dim) {
	return	(x == 0 && (p == W || p == SW || p == NW)) ||
			(x == dim->x-1 && (p == E || p == SE || p == NE)) ||
			(y == 0 && (p == N1  || p == N2 || p == NW || p == NE)) ||
			(y == dim->y-1 && (p == S1 || p == S2  || p == SW || p == SE));
}
int is_ok(int x, int y, port p, ResourceList* frs, const Params* dim) {
	return x >=0 && y >= 0 && x < dim->x && y < dim->y && !is_faulty_resource(Resource(x,y,p), frs) && !is_border_channel(x,y,p,dim);
}


// Subnetwork 1 has eastern going traffic, subnetwork 2 western.
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<10>;
    Resource *nexthops = hops->routed;
	int i = 0;

	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;
	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// Adaptive routing in NePA architecture
		if (dx < nx || c.p == W || c.p == S2 || c.p == N2 || c.p == SW || c.p == NW || c.p == L) {
			if (dx < nx && dy > ny) {
				if (is_ok(nx,ny,W,frs,dim) && 0 < nx-2 && c.p != E)
					nexthops[i++] = Resource(nx, ny, W);
				if (is_ok(nx,ny,S2,frs,dim) && dim->y > ny+2 && c.p != E)
					nexthops[i++] = Resource(nx, ny, S2);
				if (is_ok(nx,ny,SW,frs,dim))
					nexthops[i++] = Resource(nx, ny, SW);
				if (i==0) {
					if (is_ok(nx,ny,W,frs,dim) && c.p != E)
						nexthops[i++] = Resource(nx, ny, W);
					else 
						nexthops[i++] = Resource(nx, ny, S2);
				}					
			}
			else if (dx < nx && dy < ny) {
				if (is_ok(nx,ny,W,frs,dim) && 0 < nx-2 && c.p != E)
					nexthops[i++] = Resource(nx, ny, W);
				if (is_ok(nx,ny,N2,frs,dim) && 0 < ny-2 && c.p != E)
					nexthops[i++] = Resource(nx, ny, N2);
				if (is_ok(nx,ny,NW,frs,dim))
					nexthops[i++] = Resource(nx, ny, NW);
				if (i==0) {
					if (is_ok(nx,ny,W,frs,dim) && c.p != E)
						nexthops[i++] = Resource(nx, ny, W);
					else 
						nexthops[i++] = Resource(nx, ny, N2);
				}					
			}
			else if (dx == nx && dy > ny) {
				if (is_ok(nx,ny,S2,frs,dim))
					nexthops[i++] = Resource(nx, ny, S2);
				if (is_ok(nx,ny,S1,frs,dim))
					nexthops[i++] = Resource(nx, ny, S1);
				if (i==0) {
					if (is_ok(nx,ny,W,frs,dim) && c.p != E)
						nexthops[i++] = Resource(nx, ny, W);
					if (is_ok(nx,ny,SW,frs,dim))
						nexthops[i++] = Resource(nx, ny, SW);
					if (i==0)
						nexthops[i++] = Resource(nx, ny, E);
				}
			}
			else if (dx == nx && dy < ny) {
				if (is_ok(nx,ny,N2,frs,dim))
					nexthops[i++] = Resource(nx, ny, N2);
				if (is_ok(nx,ny,N1,frs,dim))
					nexthops[i++] = Resource(nx, ny, N1);
				if (i==0) {
					if (is_ok(nx,ny,W,frs,dim) && c.p != E)
						nexthops[i++] = Resource(nx, ny, W);
					if (is_ok(nx,ny,NW,frs,dim))
						nexthops[i++] = Resource(nx, ny, NW);
					if (i==0)
						nexthops[i++] = Resource(nx, ny, E);
				}
			}
			else if (dx < nx && dy == ny) {
				if (is_ok(nx,ny,W,frs,dim))
					nexthops[i++] = Resource(nx, ny, W);
				else {
					if (is_ok(nx,ny,NW,frs,dim))
						nexthops[i++] = Resource(nx, ny, NW);
					if (is_ok(nx,ny,SW,frs,dim))
						nexthops[i++] = Resource(nx, ny, SW);
					if (!is_ok(nx,ny,NW,frs,dim) && is_ok(nx,ny,N2,frs,dim) && (c.p != S1 && c.p != S2))
						nexthops[i++] = Resource(nx, ny, N2);
					if (!is_ok(nx,ny,SW,frs,dim) && is_ok(nx,ny,S2,frs,dim) && (c.p != N1 && c.p != N2))
						nexthops[i++] = Resource(nx, ny, S2);
				}
			}
		}


		if (dx > nx || c.p == E || c.p == S1 || c.p == N1 || c.p == SE || c.p == NE) {
			if (dx > nx && dy > ny) {
				if (is_ok(nx,ny,E,frs,dim) && dim->x > nx+2 && c.p != W)
					nexthops[i++] = Resource(nx, ny, E);
				if (is_ok(nx,ny,S1,frs,dim) && dim->y > ny+2 && c.p != W)
					nexthops[i++] = Resource(nx, ny, S1);
				if (is_ok(nx,ny,SE,frs,dim))
					nexthops[i++] = Resource(nx, ny, SE);
				if (i==0) {
					if (is_ok(nx,ny,E,frs,dim) && c.p != W)
						nexthops[i++] = Resource(nx, ny, E);
					else 
						nexthops[i++] = Resource(nx, ny, S1);
				}					
			}
			else if (dx > nx && dy < ny) {
				if (is_ok(nx,ny,E,frs,dim) && dim->x > nx+2 && c.p != W)
					nexthops[i++] = Resource(nx, ny, E);
				if (is_ok(nx,ny,N1,frs,dim) && 0 < ny-2 && c.p != W)
					nexthops[i++] = Resource(nx, ny, N1);
				if (is_ok(nx,ny,NE,frs,dim))
					nexthops[i++] = Resource(nx, ny, NE);
				if (i==0) {
					if (is_ok(nx,ny,E,frs,dim) && c.p != W)
						nexthops[i++] = Resource(nx, ny, E);
					else 
						nexthops[i++] = Resource(nx, ny, N1);
				}					
			}
			else if (dx == nx && dy > ny) {
				if (is_ok(nx,ny,S1,frs,dim))
					nexthops[i++] = Resource(nx, ny, S1);
				if (is_ok(nx,ny,S2,frs,dim))
					nexthops[i++] = Resource(nx, ny, S2);
				if (i==0) {
					if (is_ok(nx,ny,E,frs,dim) && c.p != W)
						nexthops[i++] = Resource(nx, ny, E);
					if (is_ok(nx,ny,SE,frs,dim))
						nexthops[i++] = Resource(nx, ny, SE);
					if (i==0)
						nexthops[i++] = Resource(nx, ny, W);
				}
			}
			else if (dx == nx && dy < ny) {
				if (is_ok(nx,ny,N1,frs,dim))
					nexthops[i++] = Resource(nx, ny, N1);
				if (is_ok(nx,ny,N2,frs,dim))
					nexthops[i++] = Resource(nx, ny, N2);
				if (i==0) {
					if (is_ok(nx,ny,E,frs,dim) && c.p != W)
						nexthops[i++] = Resource(nx, ny, E);
					if (is_ok(nx,ny,NE,frs,dim))
						nexthops[i++] = Resource(nx, ny, NE);
					if (i==0)
						nexthops[i++] = Resource(nx, ny, W);
				}
			}
			else if (dx > nx && dy == ny) {
				if (is_ok(nx,ny,E,frs,dim) && c.p != W)
					nexthops[i++] = Resource(nx, ny, E);
				else {
					if (is_ok(nx,ny,NE,frs,dim))
						nexthops[i++] = Resource(nx, ny, NE);
					if (is_ok(nx,ny,SE,frs,dim))
						nexthops[i++] = Resource(nx, ny, SE);
					if (!is_ok(nx,ny,NE,frs,dim) && is_ok(nx,ny,N1,frs,dim) && (c.p != S1 && c.p != S2))
						nexthops[i++] = Resource(nx, ny, N1);
					if (!is_ok(nx,ny,SE,frs,dim) && is_ok(nx,ny,S1,frs,dim) && (c.p != N1 && c.p != N2))
						nexthops[i++] = Resource(nx, ny, S1);
				}
			}
		}
	return hops;
}



