/**************/
/* ROUTING    */
/**************/

/*
Shortest path routing in the Spidergon ring
*/

#include "spidergon_definitions.h"


/*
The routing function. Routes from channel c == (x,p) to processing node n == (dx)
Returns a set of resources. As routing is deterministic, one next hops are returned.
This list is closed with a NULL.
*/
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<2>;
    Resource *nexthops = hops->routed;

	// The coordinates of the destination
	int dest = d->s;
	// The coordinates of the processing node at the end of channel c
	int next = get_end(c).s;

	// Compute relAd:
	int relAd = (dest - next + NUM_OF_PROC_NODES) % NUM_OF_PROC_NODES;

	if (relAd == 0) {
		// No next hop
	}
	else if (relAd == 1 || relAd == 2)
		nexthops[0] = Resource(next, CW);
	else if (relAd == 6 || relAd == 7)
		nexthops[0] = Resource(next, CCW);
	else
		nexthops[0] = Resource(next, ACC);

	return hops;
}

